/*  T-Zero Battle bot  control program May 2 2002
Ver 1.0

By Dale Heatherington, WA4DSY
dale.h@wa4dsy.net

Atmel AVR AT90S8535 microcontroller

Compiled with avr-gcc  (See Makefile)

----------------------------------------
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA

----------------------------------------
*/
 

/***************************************************************************/

//Note: CPU type defined in Makefile

#include <io.h>
#include <sig-avr.h>
#include <interrupt.h>
#include <pgmspace.h>
#include <progmem.h>
#include <eeprom.h>

#define UCHAR uint8_t
                
#define F_CPU    7372800        /* crystal frequency of CPU */

#define TRUE 1
#define FALSE 0
#define SYNC1 0xf0
#define SYNC2 0x55
#define MODE_STOP 0
#define REV 0
#define FWD 1

/* Packet: 0xff 0xf0 0x55 ADDRESS COUNT DATA... CRCH CRCL */

#define COUNT 1      /* offset to count field in tx/rx buffers */
#define ADDRESS 0    /* offset to address field in tx/rx buffers */
#define DATA 2       /* offset to start of data in tx/rx buffers */
#define RIGHTJOY_Y 3  //offset to right joystick Y value
#define RIGHTJOY_X 4  //offset to right joystick X value
#define LEFTJOY_Y 5
#define LEFTJOY_X 6
#define MATRIX0 7     //offset to matrix 0 byte  (These are the controller buttons)
#define MATRIX1 8     //offset to matrix 1 byte
#define BOT_ADDR 1   /* Address of this robot */

//Bits in Port D
#define LED_RED 2         /* low true */
#define LED_YELLOW 3
#define LED_GREEN 4
#define LED_BLUE 7
#define LED_IR_LEFT 5     /* High true */
#define LED_IR_RIGHT 6    /* High true */


/* Switch Matrix bit defs (buttons on controller)*/

/*matrix0*/

#define m_mode 0x80
#define m_east 0x40
#define m_north 0x20
#define m_rumble 0x10
#define m_south 0x08
#define m_s 0x04
#define m_left_fire 0x02
#define m_west  0x01

/*matrix1*/

#define m_right_fire 0x80
#define m_A          0x40
#define m_B          0x20
#define m_C          0x10
#define m_Z          0x08
#define m_Y          0x04
#define m_X          0x02


/*--- Motor control variables and defs ----- */

/* Port B motor and flipper control bits */
#define  FwdLeft  2     /*Left Motor forward */
#define  RevLeft  3     /*Left Motor reverse*/
#define  FwdRight 0     /*Right Motor forward */
#define  RevRight 1     /*Right Motor reverse */
#define FLIPPER   4     /* Flipper weapon firing bit (high true) */
#define  P_width  1    /*Motor pulse width in interrupt ticks (104 uS per tick ) */
UCHAR  workB;
UCHAR r_speed;
UCHAR l_speed;
UCHAR l_integrate;
UCHAR r_integrate;
UCHAR dir_ctl;
UCHAR  speed;
UCHAR  right_speed,right_prm;
UCHAR  left_speed,left_prm;
UCHAR  dir_flag_left, dir_flag_right;
UCHAR bot_mode, testmode;
UCHAR timer_led_green;
UCHAR inverted, timer_inv;
UCHAR rmfb,lmfb;    //Motor feedback values from AD conv. , range is 92 - 164 with 128=stopped
                     //lower values equal FORWARD, higher are REVERSE.

#define AD_RM_SPEED 4
#define AD_LM_SPEED 5


void adjust_motor_speed();
UCHAR speed_to_pulse_rate(UCHAR speed);
UCHAR integrate(UCHAR speed,UCHAR mfb,UCHAR integ);

/* ------- End of motor control -------------*/


UCHAR matrix0, matrix1, last_matrix0, last_matrix1, test;
UCHAR ad_m_current, ad_lm_speed, ad_rm_speed, ad_bat_volt, ad_floor_sense, ad_rssi;   //A-D converter values


#define RXSIZE 20
UCHAR tx_buf[32];      //Serial tx buffer for debug telemetry
UCHAR rx_buf[RXSIZE];     //Serial receive buffer for RC commands
UCHAR rx_pkt[RXSIZE-2];
UCHAR tx_ptr,tx_ctr , rx_ptr, rx_ctr;
UCHAR txstate,rxstate;
uint16_t crc, rxcrc,last_rx;
UCHAR  counter,val,rssi;
UCHAR floor_sens_left, floor_sens_right;
UCHAR x_offset, y_offset;
UCHAR timer_fire_flipper, timer_inhibit_flipper;

int ii;
UCHAR prescale ;
UCHAR prescale100;
UCHAR prescale1sec;
UCHAR prescaleMinute;
UCHAR timer10ms ;
UCHAR timer1sec;
unsigned int minutes;
UCHAR run_seconds;

void make_info_pkt();
void set_tx_buffer_crc();
uint16_t get_adc(uint8_t chan);
uint16_t calc_crc(char data, uint16_t crc);
void make_info_pkt();
void delay_100us(uint8_t count);
//-------------------------------------------------------------------------

UCHAR hard_limit(UCHAR x, UCHAR lim)
{

   if(x < 0x80){
      if(x > lim) x = lim;
      return x;
   }

   if(x >= 0x80){
      lim = (lim ^ 0xff) + 1;
      if(x < lim) x = lim;
      return x;
   }



}

//--------------------------------------------------------------------------
/* Read the 10 bit analog to digital converter */

uint16_t get_adc(UCHAR chan)
{  
      
    outp(chan,ADMUX);               //Select channel
    sbi(ADCSR,ADSC);                //Start conversion 
    while( bit_is_clear(ADCSR,ADIF)); //Wait for ADC output registers to be loaded
    return __inw_atomic(ADCL);       //Return 16 bit value
}

//---------------------------------------------------------------------------

/* Read 10 bit AD converter but return 
   only 8 bits discarding low 2 bits. */

uint8_t get_adc8(UCHAR chan)
{  
    uint16_t result;

    outp(chan,ADMUX);               //Select channel
    sbi(ADCSR,ADSC);                //Start conversion 
    while( bit_is_clear(ADCSR,ADIF)); //Wait for ADC output registers to be loaded
    result = __inw_atomic(ADCL);     //Get 10 bit AD value 
    return result >> 2;     //Return 8 bit value
}
//---------------------------------------------------------------------------

   // compute 16 bit CRC on 1 data byte. 
   //   crc is a  variable which accumulates the 16 bit CRC. 
   //   It must be cleared before starting to compute 16 bit CRC on a string.
   //   The output value will be returned as a 16 bit int.
   
uint16_t calc_crc(char data, uint16_t crc){

      uint8_t i;
     
      for (i=0;i<8;i++) {                 //Do all 8 bits in the data byte
         if (data & 0x80) crc ^= 0x8000;  //if data bit 7 is set then flip crc msb
         data = data << 1;                // get next data bit
         if (crc & 0x8000) {              //if crc msb is 1...
             crc = crc << 1;              // ..shift left 1
             crc = crc ^ 0x8005;          // ..xor with 0x8005
          } else crc = crc << 1;          //else just shift left 1 with no xor
        
      }

      return crc;

}   

//-----------------------------------------------------------------------
//Compute CRC on the TX packet and put it on the end
void set_tx_buffer_crc(){
      UCHAR j = tx_buf[COUNT] + 2; //get data size
      UCHAR i;

      crc = 0;
      for (i=0;i<j;i++) crc = calc_crc(tx_buf[i],crc); //Compute crc16 over buffer
      tx_buf[i++] = (uint8_t)(crc >> 8);  //tack the crc bytes on the end of the buffer
      tx_buf[i++] = (uint8_t)(crc & 0xff);
      
}
//--------------------------------------------------------------------------
//Check CRC on a received packet.  If result is zero CRC is good.
uint16_t ck_rx_crc(){

   uint16_t rx_crc;
   UCHAR j = rx_buf[COUNT] + 4; //get size of complete packet including crc bytes
   UCHAR i;

   rx_crc = 0;
   for(i=0; i < j; i++) rx_crc = calc_crc(rx_buf[i],rx_crc);  //compute CRC over complete buffer
   return rx_crc;

}
        
//---------------------------------------------------------------------------

 int main(void)
{
  UCHAR x,y;

  outp(BV(CS01), TCCR0);      /* use CLK/8 source for counter */
  
  
  outp(0x9f, PORTD);          /* turn all leds off  */
  outp(0xff, DDRD);           /* port D all outputs */

  outp(0x00,PORTB);           /* port B all bits off */
  outp(0x1f,DDRB);            /* low 5 bits outputs, motor and flipper control */

  outp(0x00,DDRA);            // 8 bits analog inputs
  outp(0,PORTA);              // Clear port A
  outp(0x85,ADCSR);           // Enable A to D converter with /32 conv clk.

  
  outp(0x00,PORTC);           // Clear port C bits
  outp(0xfe,DDRC);            // bit 0 input on port C  (inversion switch)

  outp(0x0b,TCCR2);           // clear on compare, Prescale = 32 on T/C 2  (230400 hz)
  outp(24,OCR2);              // T/C 2 compare register for 9600 hz interrupts
  sbi(TIMSK,OCIE2);           // Enable timer 2 compare interrupts
  
  outp(47,UBRR);              /* UART baud rate set to 9600 */
  sbi(UCR,RXEN);              // enable uart receiver
  sbi(UCR,RXCIE);             // receiver interupt enabled

  x_offset = 0;
  y_offset = 0;
  inverted = FALSE;
  rssi = 0;
  timer_inhibit_flipper = 100;
  timer_fire_flipper = 0;
  minutes = 0;
  run_seconds = 0;
  txstate = 0;
  l_speed = 0;                //Motors set to zero speed
  r_speed = 0;

  //bot_mode = MODE_STOP;
  testmode = 0;

  sei();                      /* enable global interrupts */

  
  for(;;){                   // ** MAIN LOOP starts here **

    /* If inversion detector switch changes state 
       for 500ms or more flip the "inverted" variable */
    if(inverted){
       if(bit_is_set(PINC,0)) timer_inv = 50; //set to 500ms timeout as long as we are inverted
       if(timer_inv == 0) inverted = FALSE;   //Timeout means we are upright again
    }else{
       if(bit_is_clear(PINC,0)) timer_inv = 50; //set to 500ms as long as we are upright
       if(timer_inv == 0) inverted = TRUE;      //timeout means we are now inverted
    }

        
    /* Get A-D converter values */
     
    ad_bat_volt    = get_adc8(3) ;
    ad_m_current   = get_adc8(6) ;
    ad_rssi        = get_adc8(7) ;


    if(ad_bat_volt < 137){
         cbi(PORTD,LED_RED);  //RED LED on if battery less than 27 volts
    }else{
       sbi(PORTD,LED_RED);
    }

            
    if(txstate == 0) make_info_pkt();     //Transmit debug telemetry

    if(timer_led_green == 0) sbi(PORTD,LED_GREEN); //Clear the green DATA LED if timeout

    if(ad_rssi > 0x48)
       cbi(PORTD,LED_YELLOW);      // DCD LED on (RF carrier detected)
       else
          sbi(PORTD,LED_YELLOW);

    y = rx_pkt[RIGHTJOY_Y] + 128 ;   //Get joystick values and convert to signed
    x = (rx_pkt[RIGHTJOY_X] >> 1) + 192;    // X joy stick is divided by 2
    
    y = y - y_offset;
    x = x - x_offset;

    if(inverted){
       y = (y ^ 0xff) + 1 ;      //Twos compliment Y joystick value if inverted
       
    }

    y = hard_limit(y,0x50);       //Limit range to +/- 0x50
    x = hard_limit(x,0x50);
    



    l_speed = hard_limit(y - x, 0x50);
    r_speed = hard_limit(y + x, 0x50);



    /* Do dead zone logic  +/- 6  */

    if((l_speed > 0xfa) && (l_speed <= 0xff)) l_speed = 0;   //Create dead zones
    else
      if((l_speed > 0) && (l_speed <= 0x6)) l_speed = 0;

    if((r_speed > 0xfa) && (r_speed <= 0xff)) r_speed = 0;
    else
      if((r_speed > 0) && (r_speed <= 0x6)) r_speed = 0;

    if(r_speed){
       if(r_speed > 128) r_speed += 6;       //Remove dead zone offsets
       else r_speed -= 6;
    }

    if(l_speed){
       if(l_speed > 128) l_speed += 6;
       else l_speed -= 6;
    }


    /* Get controller switch matrix */

    if((rx_pkt[MATRIX1] & m_Z)      //If Z button pushed...
       && (timer_led_green > 0)){   //...and valid RC data then
       y_offset = rx_pkt[RIGHTJOY_Y] + 128;      //save the joystick neutral settings
       x_offset = (rx_pkt[RIGHTJOY_X] >> 1) + 192 ;


    }


    adjust_motor_speed();

    matrix0 = rx_pkt[MATRIX0];
    if((matrix0 & m_left_fire)         //If button is pushed... 
         && ((last_matrix0 & m_left_fire) == 0) //..and was not pushed last time...
         && (timer_led_green > 0)){    //...and valid RC data present then fire the weapon
       timer_fire_flipper = 6;         //Set for 60ms valve open time.
    }

    if (((matrix0 & m_left_fire) == 0)
       && (last_matrix0 & m_left_fire)){
         timer_inhibit_flipper = 100;     //Inhibit re-firing for 1 second
    }

    last_matrix0 =  matrix0;


  }

}

//---------------------------------------


void delay_100us(uint8_t count)
{
  while(count){
    outp(256 - ((F_CPU/80000) -1), TCNT0);
    outp(BV(TOV0), TIFR);
    while(!(inp(TIFR) & BV(TOV0)));   
    count--;
  }
}


void delay_10ms(uint8_t count)
{
  while(count){
    delay_100us(100);
    count--;
  }
}

//----------------------------------------------------------------------
UCHAR integrate(UCHAR speed, UCHAR mfb, UCHAR integ){

      speed = speed + 128;   //convert to unsigned
      mfb = mfb + 128;
      if (speed > mfb) {
         integ++;
         if (integ == 0x80) integ--;  //Clip at +127
         return integ;
      } else {
         integ--;
         if (integ == 0x7f) integ++;  //Clip at -128
         return integ;
      }

}



//----------------------------------------------------------------------
   
/* Adjust motor speeds. Cause feedback values to match
   speed control input values */
  
void adjust_motor_speed(){
      UCHAR i;
      

      rmfb = get_adc8(AD_RM_SPEED) ; //get right motor feedback  (range is +/- 36 decimal)
      rmfb = (rmfb - 64) << 1  ;     //Multiply by 2   (to match joystick range)
      rmfb += 128;                  //Convert to signed values
      lmfb = get_adc8(AD_LM_SPEED) ; //get left motor feedback 
      lmfb = (lmfb - 64) << 1;       //multiply by 2
      lmfb += 128;                   //Convert to signed

      for (i=0;i<3;i++) {  //Adjust motor speeds based on feedback values
     
        r_integrate = integrate(r_speed,rmfb,r_integrate);
        l_integrate  = integrate(l_speed,lmfb,l_integrate);
        
       
      }
      right_speed = r_integrate;
      left_speed  = l_integrate;
 }


//-----------------------------------------------------------------------
   /* Convert a signed value to an absolute value between 127 and 0 
        0 becomes 127
      127 becomes 0
     -128 becomes 0
     
   */
UCHAR speed_to_pulse_rate(UCHAR speed)
   {
      if (speed > 127) {      //See if speed is negative !
         return speed & 0x7f;
      } else {
         return speed ^ 0x7f;
      }



   }

   

//-----------------------------------------------------------------------
/* Come here 9600 times per second */
SIGNAL(SIG_OUTPUT_COMPARE2)
{
   
      if(--prescale == 0){      //This executes 100 times per sec.
         prescale = 96;
         if(prescale1sec > 0) prescale1sec--;
         if(timer_led_green > 0) timer_led_green--;
         if(timer_inv > 0) timer_inv--;
         if(timer_fire_flipper > 0) timer_fire_flipper--;
         if(timer_inhibit_flipper > 0) timer_inhibit_flipper--;
         
      }

      
     if(prescale1sec == 0){  //This executes once per second
        prescale1sec = 100;
        if(prescaleMinute > 0) prescaleMinute--;
        run_seconds++;
        
     }

     if(prescaleMinute == 0){ //This executes once per minute
        prescaleMinute = 60;
        minutes++;
        
        
     }


 /* Motor control code */
   
     outp(workB,PORTB);     //Output to the H-Bridge
     
     workB = 0;
      

         /* adjust right pulse rate modulator */
         if (right_prm < P_width) {
            if (dir_flag_right == FWD)
               workB |= BV(FwdRight);   //Set FwdRight bit 
            else
               workB |= BV(RevRight);   //else set RevRight bit
         }
         if (right_prm-- == 0) {
            right_prm = speed_to_pulse_rate(right_speed) + P_width -1;
            if (right_speed > 127) dir_flag_right = REV;
               else dir_flag_right = FWD;
         }

         /*adjust left pulse rate modulator */
         if (left_prm < P_width) {
            if (dir_flag_left == FWD)
               workB = workB | BV(FwdLeft) ;
            else
               workB = workB | BV(RevLeft);
         }

         if (left_prm-- == 0) {
            left_prm = speed_to_pulse_rate(left_speed) + P_width -1;
            if (left_speed > 127) dir_flag_left = REV;
            else dir_flag_left = FWD;
         }


      

    if (ad_m_current > 20) workB = 0;  // Limit total motor current to 4 amps
    //if (bot_mode == MODE_STOP) workB = 0;    //Stop the bot.
      
    if ((l_speed == 0) && (r_speed == 0)) workB = 0;
    if(testmode == 1) workB = 0;
    if(timer_led_green == 0) workB = 0;  //Stop the bot if no control data

    if((timer_fire_flipper > 0)
       && (timer_inhibit_flipper == 0)) workB |= BV(FLIPPER);   //Fire flipper weapon

    /* End motor control code */


}

//-----------------------------------------------------------------------
/* UART receive interrupt handler */

SIGNAL(SIG_UART_RECV)
{
   static UCHAR rxstate = 0;
 
    UCHAR ch,i;
    
    ch = inp(UDR);      //Get the current rx char
    

    switch (rxstate){
 
    case 0:    if((ch == SYNC2) && (last_rx == SYNC1)){
                   rxstate = 1;
                   
               }
               break;


    case 1:    rx_buf[ADDRESS] = ch;
               rxstate = 2;
               break;

    case 2:    rx_buf[COUNT] = ch;
               rxstate = 3;
               rx_ptr = DATA;
               rx_ctr = ch + 2;
               if((rx_ctr + 4) > RXSIZE) rxstate = 0;  //Quit if buffer would overflow
               break;

    case 3:    rx_buf[rx_ptr++] = ch;
               rx_ctr--;
               if(rx_ctr == 0){
                  if((ck_rx_crc() == 0) && (rx_buf[ADDRESS] == BOT_ADDR)){     //Good packet received for us
                     for(i=0; i < (rx_buf[COUNT] + 2); i++) rx_pkt[i] = rx_buf[i];  //Copy to packet buffer
                     cbi(PORTD,LED_GREEN);  // Light up the data LED
                     timer_led_green = 5;   //50ms timeout
                  }

                  rxstate = 0;

               }

                   
               break;

               
    }


   last_rx = ch;       //Save last char

}

 //-----------------------------------------------------------------------

   /* Build a packet and send debug data */
  
   void make_info_pkt(){
      uint8_t i;
      i = ADDRESS;
      tx_buf[i] = 1;             // address of monitor
      i = DATA;
      tx_buf[i++] = 0x01;        //data type = 1
      tx_buf[i++] = l_speed;           //1  Left motor speed
      tx_buf[i++] = r_speed;           //2  right Motor speed
      tx_buf[i++] = ad_m_current;      //3  Motor current
      tx_buf[i++] = ad_bat_volt;       //4  Battery voltage/10
      tx_buf[i++] = ad_rssi;           //5   recv sig strength
      tx_buf[i++] = floor_sens_left;   //6  left floor sensor
      tx_buf[i++] = floor_sens_right;  //7  right floor sensor
      tx_buf[i++] = matrix0;          //8   controller switch matrix byte 0
      tx_buf[i++] = matrix1;          //9   controller switch matrix byte 1
      tx_buf[i++] = run_seconds;       //10 run time in seconds
      tx_buf[i++] = lmfb;
      tx_buf[i++] = rmfb;
            
      tx_buf[COUNT] = i - DATA;       //Set byte count field to data length

      set_tx_buffer_crc();  //compute and add CRC16 bits to end of pkt
      
      txstate = 2;          //Tell scheduler it's ready to transmit
      //outp(0x28,UCR);       //enable UART TX and interrupt 
      sbi(UCR, TXEN); 
      sbi(UCR, UDRIE);
      
      outp(0xff,UDR);       // Send a FF to get things started 
   }


//----------------------------------------------------------------------


/* UART Transmit interrupt handler */  
SIGNAL(SIG_UART_DATA)
{ 
   

      switch(txstate) {
      case 0:  break;
      case 1:
      case 2:
      case 3:  outp(0xff,UDR) ; txstate++;
         break;
      case 4:  outp(SYNC1,UDR); txstate++;
         break;
      case 5:  outp(SYNC2,UDR); txstate++;
         break;

      case 6:  tx_ptr = 0;
         tx_ctr = tx_buf[COUNT] + 4; 
         txstate++;
         //fall through...

      case 7:  outp(tx_buf[tx_ptr++],UDR);  //Now send buffer contents
            tx_ctr--;
            if (tx_ctr == 0){ 
               txstate = 8;
               
            }
         
            break;

      case 8:   
               txstate = 0;
               outp(0xff,UDR);   //Feed the uart data register one last time to clear this interrupt
               cbi(UCR,UDRIE);   //Stop transmitter and interrupts
               cbi(UCR,TXEN);
               break;



      }

      
}


//---------------------------------------------------------------------------

/* end of file */



